{**********************************************************************

                          SAMPLE CONVERSION PROGRAM

***********************************************************************

  Carter coordinate to Latitude/Longitude conversion example

  Program by: Brandon C. Nuttall

              Kentucky Geological Survey
              228 Mining & Mineral Resources Building
              University of Kentucky
              Lexington, KY  40506-0107

  Language:   Borland International TURBO Pascal v5.0
  Computer:   DEC Rainbow PC 100B (512K)
              IBM PC AT & XT clones
  Operating system: MS-DOS v3.0 or greater

**************************************************************************
*                                                                        *
*                            N O T I C E                                 *
*                                                                        *
*     This software is placed in the public domain by the Kentucky       *
*     Geological Survey and is supplied "as is".                         *
*                                                                        *
*     The Kentucky Geological Survey makes no warranty either express    *
*     or implied as to the accuracy or validity of these routines.  The  *
*     user assumes all liability for any losses or damages that may be   *
*     incurred through any use or application of this software.          *
*                                                                        *
**************************************************************************

}

program Sample(input,output);

{
  3-3-89 v1.1 BCN
    1) Use Crt unit for screen control
    2) Fix runtime error 103, trying to close an unopend file
    3) Add YesOrNo function to handle appropriate questions
    4) Add optional output of lat/lon to an ASCII file appropriate
       for editing or import into other software

  3-23-89 v1.2 BCN
    1) Add feature to allow use of escape key to terminate program
       execution

  6-21-89 v2.0 BCN
    1) Integrate latitude/longitude to Carter coordinate conversion

  6-23-89      BCN
    Fix bug in detection of existing output file.  Bad parameter equal
    to the input file spec was supplied thus the output file was always
    found to exist.

  6-01-90 v2.1 BCN
    Allow decimal degrees and degrees minutes seconds input of latitude
    and longitude values.
    1) add charset data type
    2) add parse function
    3) add StrToDD function and put it in AskReal() and MakeReal()

  7-08-94 v3.0 BCN
    Recompile to take advantage of changes to Carterll unit that result
    in ignoring tab characters in input.
    Revise "Illegal integer input" messages to be more informative
    Set MINSTR to 9 (was 17) to allow id and carters like:

         1 5n4e1b9

    i.e., id = 1 and carter = 0005 fnl 0004 fel 1-b-9

}

Uses
  Crt, Carterll;

const
  pgmname = 'Sample Coordinate Conversion Demonstration Program v2.1';
  EOS     =   #0; { use ASCII null character to terminate long strings  }
  MINSTR  =    9; { (I don't see how a Carter coordinate and id can be
                    fewer characters than this if written in printable
                    ASCII characters) }
  MAXSTR  = 2048; { this is the current maximum record length supported }
  STRLEN  = MAXSTR + 1; { length to hold a terminating character        }

type
  longstr = array[1..STRLEN] of char; { this is the basic data structure }
  charset = set of char;

var
  data     : longstr;
  datafile : text;    { assume a file of ASCII text, each record terminated
                        by a carriage return }
  out      : text;    { output an ASCII text file if requested }
  lrecl    : integer; { user data file logical record length }
  lenid    : integer; { length of "unique" record identifier }
  lencc    : integer; { length of the Carter coordinate string }
  ccoffset : integer; { offset from beginning of record to the Carter
                        coordinate string }
  idoffset : integer; { offset from beginning of record to the unique
                        identifier string }
  id,cc    : string;              { the identifier and Carter coordinate }
  lat,lon  : real;                { the latitude and longitude }
  latoffset,          { offset from beginning of record to latitude string }
  lonoffset,          { offset from beginning of record to longitude string }
  lenlat,             { length of latitude string }
  lenlon   : integer; { length of longitude string }
  InFile,                         { file name for disk input }
  OutFile  : string;              { file name for disk output }
  pause,                          { pause on error ? }
  CCtoLL,                         { true for Carter to lat/lon conversion }
  Keyboard,                       { true if input from keyboard }
  Save,                           { true if saving lat/lons to file }
  Again    : boolean;             { false if user quits or on end of file }
  count    : integer;             { provide pseudo-id for keyboard input }

Function YesOrNo(Msg : string) : boolean;
var
  ch : char;
begin
  write(Msg,' (y/n)? ');
  repeat
    ch := readkey;
    if not (upcase(ch) in ['Y','N']) then write(^G);
  until upcase(ch) in ['Y','N'];
  writeln(ch);
  YesOrNo:=(upcase(ch)='Y')
end;

Procedure ESCPressed;
begin
  writeln;
  if YesOrNo('Do you really want to quit') then begin
    if not keyboard then close(datafile); { close input file  }
    if Save then close(out);              { close output file }
    writeln; writeln('Bye');
    halt;
  end;
end;

Procedure PressAnyKey;
var
  ch : char;
begin
  writeln;
  writeln('Press any key to continue or press ESC to quit');
  repeat until keypressed;
  ch := readkey;
  if ch=#0 then ch := readkey; { eat extended scan code }
  if ch=#27 then ESCPressed;
  writeln;
end;

Procedure CheckHalt;
var
  ch : char;
begin
  if keypressed then begin { don't do anything if a key is not pressed }
    ch := readkey; { get the key }
    if ch=#0 then ch := readkey; { eat extended scan code of function keys }
    if ch=#27 then ESCPressed;
  end;
end;

Procedure AskInt(Prompt : string; Var Param: Integer;
                Low, High: Integer; EMsg : string);
  Var
    ok     : boolean;
    Temp   : string;
    Legal  : Integer;
  Begin
    repeat
      write(Prompt);
      readln(Temp);
      Val(Temp,Param,Legal);
      ok := (legal=0);
      if ok then ok := (Param >= Low) and (Param <= High);
      if not ok then begin
        writeln('Illegal integer input');
        writeln(EMsg);
      end;
      CheckHalt; { BCN 8-Jul-1994 allow bail out here also }
    until ok
  End;

function parse(var str, token : string;
                         dlm : charset;
                var  numeric : boolean) : boolean;
{
  Parse "peals" all characters not in the set of delimiters, dlm,
    from the input string, str, up to the first occurrance of any
    character in the set of delimiters

  The string of characters found is returned as the value of "token"

  If there is a token to be found, parse succeeds, returns true,
    otherwise, it returns false.  The failure of parse usually
    indicates the end of the input has been reached.

  Parse currently recognizes two basic data types
    1) a span of the digits 0 to 9 possibly including "+", "-", ".", and
       'e' is numeric.
    2) everything else is not
}
var
  i : byte;

  function is_numeric(token : string) : boolean;
  var
    err : integer;
    v   : real;
  begin
    val(token,v,err); { cheat, if val() can make it a number, then it is
                        numeric }
    is_numeric := (err=0);
  end;

begin { parse }
  token := '';                            { initialize token }
  while (str<>'') and
        (str[1] in dlm) do delete(str,1,1); { strip leading delimiters }
  if length(str)>0 then begin             { if something remains... }
    repeat
      token := token + str[1];            { add it to the output token }
      delete(str,1,1)                     { delete it from the input str }
    until (str[1] in dlm) or (length(str)=0); { stop at a delimiter or
                                                at the end of the input }
    parse := true;                        { we got something }
    numeric := is_numeric(token);         { is it a number? }
  end else begin { we were looking for something and nothing was there }
    parse := false
  end;
end;

function StrToDD(s : string; var dd : real) : boolean;
const
  dlm : charset = [' ','d','m','s','D','M','S',','];
var
  t : string;
  e : integer;
  ok, num : boolean;
  d, m, sec, n : real;
begin
  dd := 0.0;
  d  := 0.0;
  m  := 0.0;
  sec:= 0.0;
  n  := 1.0; { store the sign here }
  StrToDD := false;
  { get degrees }
  ok := parse(s,t,dlm,num);
  if not (ok or num) then exit;
  val(t,d,e);
  { store sign }
  if d<0.0 then begin
    n := -1.0;
    d := abs(d);
  end;
  { if there, check on minutes }
  if s<>'' then begin
    ok := parse(s,t,dlm,num);
    if not (ok or num) then exit;
    val(t,m,e);
    if (m<0.0) or (m>=60.0) then begin
      writeln('Bad number of minutes');
      exit;
    end;
    { do seconds }
    if s<>'' then begin
      ok := parse(s,t,dlm,num);
      if not (ok or num) then exit;
      val(t,sec,e);
      if (sec<0.0) or (sec>=60.0) then begin
        writeln('Bad number of seconds');
        exit;
      end;
    end;
  end;
  { calculate decimal degrees }
  dd := ((((sec / 60.0) + m) / 60.0) + d) * n;
  StrToDD := true;
end;

procedure AskReal(Prompt : string; var Param : real);
{ prompted input of any real number }
var
  ok : boolean;
  temp : string;
  legal : integer;
begin
  repeat
    write(Prompt);
    readln(Temp);
    Val(Temp,Param,Legal);
    ok := (legal=0);
    if not ok then begin
      ok := StrToDD(temp,param); { check sexagesimal input }
      if not ok then writeln('Illegal real number input');
    end;
  until ok
End;

function Exists(FileSpec : string) : boolean;
{ see if a file of the given name exists }
  var
    DoNotCare : file; { use untyped file here, we don't really care }
    succeeded : boolean;
  begin
    {$I-}
    Assign(DoNotCare,FileSpec);
    Reset(DoNotCare);
    succeeded := (IOResult=0);
    if succeeded then Close(DoNotCare);
    Exists := succeeded;
    {$I+}
  end;

procedure Introduce;
begin
  ClrScr;                         { program identification }
  writeln(pgmname);
  writeln;
  writeln(ccllsource);            { print source of routines }
  writeln(ccllname);              { print name of routine & version }
  writeln;
  writeln('This program performs conversions between the Carter ',
    'coordinate system');
  writeln('in Kentucky and latitude and longitude.  For input',
    ' requiring a yes or');
  writeln('no answer, press "Y" or "N", pressing enter is not',
    ' required.');
  writeln;
  writeln('If you have any questions or problems, please contact:');
  writeln;
  writeln('     Brandon C. Nuttall');
  writeln('     Kentucky Geological Survey');
  writeln('     228 Mining & Mineral Resources Building');
  writeln('     University of Kentucky');
  writeln('     Lexington, KY   40506-0107');
  writeln;
  writeln('     Phone: (606) 257-5500');
  PressAnyKey;
end;

procedure Initialize;
  var
    ch : char;
    ok : boolean;
  begin
    pause := false;  { don't pause on error (getting input data from file) }
    ClrScr;

    count := 0;                     { initialize pseudo-id }
    lenid := 5;                     { establish default for keyboard
                                      input }

    CCtoLL := not
      YesOrNo('Do you want to convert latitude and longitude to Carters');

    keyboard :=
      YesOrNo('Do you want to enter input data from the keyboard');

    if not CCtoLL then Expand :=
      YesOrNo('Do you want the expanded format Carter coordinate');

    if not keyboard then begin { set up for disk file input }
      repeat
        write('Enter input data file specification: ');
        readln(InFile);
        ok := Exists(InFile);
        if not ok then writeln(^G,'Can''t find file or bad file name');
      until ok;
      assign(datafile,InFile);
      reset(datafile);
      if not keyboard then
        pause := YesOrNo('Do you want to pause whenever there is an error');
      AskInt('Enter record length:                  ',lrecl,MINSTR,MAXSTR,
        'Record length must be between 9 and 2048');
      AskInt('Enter length of record identifier:    ',lenid,1,255,
        'Identifier must be at least 1 but no more than 255 characters');
      AskInt('Enter offset to start of identifier:  ',idoffset,0,lrecl-lenid,
        'Identifier cannot extend beyond end of record');
      if CCtoLL then begin
        AskInt('Enter length of Carter coordinate:    ',lencc,MINSTR,255,
          'Carter coordinate must be at least 7 but no more than 255 characters');
        AskInt('Enter offset to start of coordinates: ',ccoffset,0,
          lrecl-lencc,'Carter coordinate cannot extend beyond end of record');
      end else begin
        AskInt('Enter length of latitude:             ',lenlat,1,255,
          'Latitude should be between 1 and 255 characters');
        AskInt('Enter offset to start of latitude:    ',latoffset,1,lrecl,
          'Latitude cannot extend beyond end of record');
        AskInt('Enter length of longitude:            ',lenlon,1,255,
          'Longitude should be between 1 and 255 characters');
        AskInt('Enter offset to start of longitude:   ',lonoffset,1,lrecl,
          'Longitude cannot extend beyond end of record');
      end;
    end;

    writeln;
    Save := YesOrNo('Do you want to save data to a file');

    if Save then begin     { set up for saving data }
      repeat
        write('Enter output data file specification: ');
        readln(OutFile);
        ok := not Exists(OutFile); { things are ok here if new file }
        if not ok then  { the file exists, check before destroying it }
          ok := YesOrNo(^G+OutFile+' already exists, replace it');
      until ok; { file doesn't exist or user is replacing it }
      assign(out,OutFile);
      rewrite(out);
    end;



  end; { Initialize }

function ReadOneRecord(var data : longstr) : boolean;
{ read one record from the input file

  returns FALSE if eof(datafile) detected else returns TRUE

  note: end of record (end of one line) is expected to be a carriage
        return }
var
  i : 0..MAXSTR;
begin
  ReadOneRecord := false;
  if not eof(datafile) then begin
    i := 0;
    repeat                   { repeat for each character in record }
      i := i+1;
      read(datafile,data[i]) { read ith character }
    until eoln(datafile) or (i=MAXSTR);
    readln(datafile);        { position read pointer to start of next
                               record }
    data[i+1] := EOS;        { mark the end of the string }
    ReadOneRecord := true;
  end;
end;

function GetField(recid, name : string; offset, len : integer;
                  data : longstr; var value : string) : boolean;
var
  i : 0..MAXSTR;
begin
  GetField := true;
  value := '';
  for i := 1 to len do begin
    if data[i+offset]<>EOS
      then value := value + data[i+offset]
      else begin
        writeln('Attempt to read past end of string while reading ',name);
        if name='identifier'
          then writeln('Found "',value,'"')
          else writeln('Found "',value,'" at id ',recid);
        GetField := false;
        PressAnyKey;
      end
  end;
end;

function GetIdAndCC(var id, cc : string) : boolean;
{ get an identifier and a Carter coordinate }
var
  ok : boolean;
begin
  GetIdAndCC := true;
  if Keyboard then begin        { get input from keyboard }
    count := count + 1;
    str(count,id);
    writeln;
    write('Enter Carter coordinate or QUIT to quit: ');
    readln(cc);
    if upcase(cc[1]) = 'Q' then GetIdAndCC := false;
  end else begin                { get input from file }
    id := '';                   { initialize id and cc }
    cc := '';
    repeat
      if ReadOneRecord(data) then begin { get a record }
        ok := GetField('','identifier',idoffset,lenid,data,id);
        if ok then
          ok := GetField(id,'Carter coordinate',ccoffset,lencc,data,cc);
      end else begin { have run out of records, return }
        GetIdAndCC := false;
        exit
      end;
    until ok;
    GetIdAndCC := ok;
  end;
end; { GetIdAndCC }

function GetLatLon(var id : string; var lat, lon : real) : boolean;
{ get an identifier and a latitude/longitude pair }
var
  ok : boolean;
  temp : string;

  function MakeReal(what, s : string; var x : real) : boolean;
  var
    legal : integer;
    ok : boolean;
  begin
    MakeReal := true;
    val(s,x,legal);
    ok := (legal=0);
    if not ok then begin
      { check for supported dms format }
      ok := StrToDD(s,x);
      if not ok then begin
        writeln('Bad input for ',what);
        writeln('Found "',s,'" at id ',id);
        MakeReal := false;
        PressAnyKey;
      end
    end
  end; { MakeReal }

begin
  GetLatLon := true;
  if Keyboard then begin        { get input from keyboard }
    count := count + 1;
    str(count,id);
    writeln;
    AskReal('Enter latitude (or 0.0 to quit):  ',lat);
    if lat > 0.0 then begin
      AskReal('Enter longitude:                  ',lon);
      GetLatLon := true;
    end else GetLatLon := false;
  end else begin             { get input from file }
    id := '';                { initialize id and lat/lon }
    lat := 0.0;
    lon := 0.0;
    repeat
      if ReadOneRecord(data) then begin
        if GetField('','identifier',idoffset,lenid,data,id) then
          if GetField(id,'latitude',latoffset,lenlat,data,temp) then
            if MakeReal('latitude',temp,lat) then
              if GetField(id,'longitude',lonoffset,lenlon,data,temp) then
                if MakeReal('longitude',temp,lon) then ok := true
                else ok := false
              else ok := false
            else ok := false
          else ok := false
        else ok := false;
      end else begin
        GetLatLon := false; { have reached end of file }
        exit
      end
    until ok;
    GetLatLon := ok;
  end;
end; { GetLatLon }

begin { main }
  Keyboard := true;               { assume keyboard input }
  Save     := false;              { assume not saving data }
  CCtoLL   := true;               { assume converting to lat/lon }
  Introduce;
  Initialize;
  writeln;
  Again := true;
  while Again do begin
    CheckHalt;
    if CCtoLL
      then Again := GetIdAndCC(id,cc)
      else Again := GetLatLon(id,lat,lon);
    { do conversion and output to console }
    if Again then begin { got some input }
      if CCtoLL then begin { convert Carter to lat/lon }
        if CarterToLatLon(id,cc,lat,lon) then begin
          writeln(id:lenid,' "',cc,'" ',lat:10:6,' ',lon:10:6);
          if save then
            writeln(out,id:lenid,' "',cc,'" ',lat:10:6,' ',lon:10:6);
          { pause display when not saving and getting input from file }
          if not (save or keyboard) then PressAnyKey;
        end else begin
          writeln(id:lenid,' "',cc,'"');
          if save then
            writeln(out,id:lenid,' "',cc,'" ',lat:10:6,' ',lon:10:6);
          if pause then PressAnyKey;
        end;
      end else begin { convert lat/lon to Carter }
        if LatLonToCarter(id,cc,lat,lon,expand) then begin
          writeln(id:lenid,' "',cc,'" ',lat:10:6,' ',lon:10:6);
          if save then
            writeln(out,id:lenid,' "',cc,'" ',lat:10:6,' ',lon:10:6);
          if not (save or keyboard) then PressAnyKey;
        end else begin
          writeln(id:lenid,' ',lat:10:6,' ',lon:10:6);
          if save then
            writeln(out,id:lenid,' "',cc,'" ',lat:10:6,' ',lon:10:6);
          if pause then PressAnyKey;
        end
      end; { conversions }
      { if not saving the data and getting input from a file, then
        stop so user can see output }
    end else begin { Again }
      if keyboard then begin
        Again := not YesOrNo('Do you really want to quit');
        if not Again then writeln('Bye');
      end;
    end;
  end;
  if not keyboard then close(datafile); { close input file  }
  if Save then close(out);              { close output file }
  writeln('Bye');
end.
